@@ -14,38 +14,25 @@ from .schem_random import SChemRandom
 from .waldo import Waldo, Instruction, InstructionType
 
 
-# Dimensions of component types
-# TODO: Refactor so default dimensions and allowable variations on type/shape are built-in to each subclass of Component
-#       E.g. Component.DEFAULT_DIMENSIONS = (2, 3), Recycler.DEFAULT_DIMENSIONS = (5, 5)
-#       Then we should be able to avoid type string manipulations ('output' in type) and only need to specify the types
-#        that change their shape from the class' default
+# Dimensions of components that differ from the standard dimensions for their type
 COMPONENT_SHAPES = {
-    # SpaceChem stores co-ordinates as col, row
+    # SC stores co-ordinates as col, row
     'research-input': (1, 1),
     'research-output': (1, 1),
     'disabled-output': (1, 1),
-    'reactor': (4, 4),
-    'output': (2, 3),  # All production level outputs appear to be 2x3
-    'drag-recycler': (5, 5),
-    'drag-storage-tank': (3, 3),
-    'drag-storage-tank-infinite': (3, 3),
-    'freeform-counter': (2, 3),
-    'drag-arbitrary-input': (2, 3),
     'drag-silo-input': (5, 5),
     'drag-atmospheric-input': (2, 2),
     'drag-oceanic-input': (2, 2),
     'drag-powerplant-input': (14, 15),
     'drag-mining-input': (3, 2),
     'drag-ancient-input': (2, 2),
-    'drag-spaceship-input': (2, 2),  # TODO: Actually (2,3) but its pipe isn't in the middle which fucks our assumptions
-    'drag-programmed-input': (2, 3),
-    'drag-qpipe-in': (3, 1),
-    'drag-qpipe-out': (3, 1),
-    'drag-printer-output': (2, 3),
-    'drag-printer-passthrough': (2, 3)}
+    'drag-spaceship-input': (2, 2)}  # TODO: Actually (2,3) but its pipe isn't in the middle which fucks our assumptions
 
-# Production level codes don't specify available reactor properties like research levels; encode them here
+# Custom and ResNet research levels use the custom research reactor type and specify all features directly in the
+# level JSON. Specify the features of all other reactor types.
+DEFAULT_RESEARCH_REACTOR_TYPE = 'custom-research-reactor'
 REACTOR_TYPES = {
+    # Standard production/research reactor types
     'drag-starter-reactor': {'bonder-count': 4},
     'drag-disassembly-reactor': {'bonder-minus-count': 4, 'has-bottom-input': False},
     'drag-assembly-reactor': {'bonder-plus-count': 4, 'has-bottom-output': False},
@@ -54,8 +41,17 @@ REACTOR_TYPES = {
     'drag-superbonder-reactor': {'bonder-count': 8},
     'drag-nuclear-reactor': {'bonder-count': 4, 'has-fuser': True, 'has-splitter': True},
     'drag-quantum-reactor': {'bonder-count': 4, 'has-sensor': True, 'has-teleporter': True,
-                             'quantum-walls-y': {5: [0, 1, 2, 3, 4, 5, 6, 7]}},
-    'drag-sandbox-reactor': {'bonder-count': 8, 'has-sensor': True, 'has-fuser': True, 'has-splitter': True}}
+                             'quantum-walls-y': {'5': [0, 1, 2, 3, 4, 5, 6, 7]}},
+    'drag-sandbox-reactor': {'bonder-count': 8, 'has-sensor': True, 'has-fuser': True, 'has-splitter': True},
+    # Reactor types that only appear in official non-ResNet research levels.
+    'empty-research-reactor': {},  # 1-1 - 1-3
+    'tutorial-research-reactor-2': {'bonder-count': 4},  # 1-4
+    'drag-reduced-reactor': {'bonder-count': 2},  # 6-1, 7-2
+    'drag-largeoutput-reactor': {'bonder-count': 4, 'has-sensor': True, 'has-large-output': True},  # 6-3, 7-4, 8-2
+    'drag-advancedfusion-reactor': {'bonder-count': 2, 'has-sensor': True, 'has-fuser': True},  # 7-3
+    'drag-quantum-reactor-x': {'has-teleporter': True, 'quantum-walls-y': {'5': [0, 1, 2, 3, 4, 5, 6, 7]}},  # QT-1
+    'drag-quantum-reactor-s': {'bonder-count': 2, 'has-sensor': True, 'has-teleporter': True,  # QT-2 - QT-4
+                               'quantum-walls-y': {'5': [0, 1, 2, 3, 4, 5, 6, 7]}}}
 
 
 class Pipe(list):
@@ -63,12 +59,12 @@ class Pipe(list):
     __slots__ = 'posns',
 
     def __init__(self, posns):
-        '''Construct a pipe. posns should be defined relative to the pipe's parent component posn.'''
+        """Construct a pipe. posns should be defined relative to the pipe's parent component posn."""
         super().__init__([None for _ in posns])
         self.posns = posns
 
     def move_contents(self):
-        '''Shift all molecules in this pipe forward one if possible.'''
+        """Shift all molecules in this pipe forward one if possible."""
         # Iterate from the back to make clog-handling simple
         for i in range(len(self) - 2, -1, -1):  # Note that we don't need to shift the last element
             if self[i + 1] is None:
@@ -76,9 +72,9 @@ class Pipe(list):
 
     @classmethod
     def from_preset_string(cls, start_posn, dirns_str):
-        '''Construct a pipe from the given CE pipe string, e.g. 'RRDRUULR', moving in the indicated directions
+        """Construct a pipe from the given CE pipe string, e.g. 'RRDRUULR', moving in the indicated directions
         (U = Up, R = Right, D = Down, L = Left) from the start_posn (should be relative to the parent component's posn).
-        '''
+        """
         posns = [start_posn]
         char_to_dirn = {'U': UP, 'R': RIGHT, 'D': DOWN, 'L': LEFT}
         for dirn_char in dirns_str:
@@ -90,9 +86,9 @@ class Pipe(list):
 
     @classmethod
     def from_export_str(cls, export_str):
-        '''Note that a pipe's solution lines might not be contiguous. It is expected that the caller filters
+        """Note that a pipe's solution lines might not be contiguous. It is expected that the caller filters
         out the lines for a single pipe and passes them as a single string to this method.
-        '''
+        """
         lines = [s for s in export_str.split('\n') if s]  # Split into non-empty lines
         # Ensure all non-empty lines are valid and for the same-indexed pipe
         assert all(s.startswith('PIPE:0,') for s in lines) or all(s.startswith('PIPE:1,') for s in lines), \
@@ -115,16 +111,16 @@ class Pipe(list):
         return Pipe(posns)
 
     def export_str(self, pipe_idx=0):
-        '''Represent this pipe in solution export string format.'''
+        """Represent this pipe in solution export string format."""
         return '\n'.join(f'PIPE:{pipe_idx},{posn.col},{posn.row}' for posn in self.posns)
 
 
 class Component:
-    '''Informal Interface class defining methods overworld objects will implement one or more of.'''
+    """Informal Interface class defining methods overworld objects will implement one or more of."""
     __slots__ = 'type', 'posn', 'dimensions', 'in_pipes', 'out_pipes'
 
     def __new__(cls, component_dict=None, _type=None, **kwargs):
-        '''Return a new object of the appropriate subclass based on the component type.'''
+        """Return a new object of the appropriate subclass based on the component type."""
         # If this is being called from a child class, behave like a normal __new__ implementation (to avoid recursion)
         if cls != Component:
             return object.__new__(cls)
@@ -161,21 +157,11 @@ class Component:
     def __init__(self, component_dict=None, _type=None, posn=None, num_in_pipes=0, num_out_pipes=0):
         self.type = _type if _type is not None else component_dict['type']
         self.posn = Position(*posn) if posn is not None else Position(col=component_dict['x'], row=component_dict['y'])
+        self.dimensions = COMPONENT_SHAPES[self.type] if self.type in COMPONENT_SHAPES else self.DEFAULT_SHAPE
+
         self.in_pipes = [None for _ in range(num_in_pipes)]
-
-        # Initialize output pipes, accounting for any level-preset pipes
+        # Initialize output pipes in middle, rounded down, accounting for any level-preset pipes
         self.out_pipes = []
-
-        type_parts = self.type.split('-')
-        if self.type in COMPONENT_SHAPES:
-            self.dimensions = COMPONENT_SHAPES[self.type]
-        elif 'output' in type_parts or 'production-target' in self.type:
-            self.dimensions = COMPONENT_SHAPES['output']
-        elif 'reactor' in type_parts:
-            self.dimensions = COMPONENT_SHAPES['reactor']
-        else:
-            raise ValueError(f"Dimensions of component {self.type} are unknown")
-
         pipe_start_posn = Position(col=self.dimensions[0], row=(self.dimensions[1] - 1) // 2)
         if component_dict is not None and 'output-pipes' in component_dict:
             assert len(component_dict['output-pipes']) == num_out_pipes, f"Unexpected number of output pipes for {self.type}"
@@ -189,7 +175,7 @@ class Component:
 
     @classmethod
     def parse_metadata(cls, s):
-        '''Given a component export string or its COMPONENT line, return its component type and posn.'''
+        """Given a component export string or its COMPONENT line, return its component type and posn."""
         component_line = s.strip('\n').split('\n', maxsplit=1)[0]  # Get first non-empty line
 
         # Parse COMPONENT line
@@ -204,9 +190,9 @@ class Component:
         return component_type, component_posn
 
     def update_from_export_str(self, export_str, update_pipes=True):
-        '''Given a matching export string, update this component. Optionally ignore pipe updates (namely necessary
+        """Given a matching export string, update this component. Optionally ignore pipe updates (namely necessary
         for Ω-Pseudoethyne which disallows mutating a 1-long pipe where custom levels do not.
-        '''
+        """
         component_line, *pipe_lines = (s for s in export_str.split('\n') if s)  # Remove empty lines and get first line
 
         _, component_posn = self.parse_metadata(component_line)
@@ -246,11 +232,11 @@ class Component:
         return f'{self.type},{self.posn}'
 
     def do_instant_actions(self, _):
-        ''''Do any instant actions (e.g. execute waldo instructions, spawn/consume molecules).'''
+        """'Do any instant actions (e.g. execute waldo instructions, spawn/consume molecules)."""
         return
 
     def move_contents(self, _):
-        '''Move the contents of this object (e.g. waldos/molecules), including its pipes.'''
+        """Move the contents of this object (e.g. waldos/molecules), including its pipes."""
         for pipe in self.out_pipes:
             pipe.move_contents()
 
@@ -263,14 +249,21 @@ class Component:
         return self
 
     def export_str(self):
-        '''Represent this component in solution export string format.'''
-        # TODO: I'm still not sure what the 4th component field is used for. Custom reactor names maybe?
+        """Represent this component in solution export string format."""
+        # By SC's convention, set the first lines of the export to be the first segment of each pipe
+        # Reverse zip to separate the first lines of each export
+        first_segment_lines, remainders = [], []
+        for i, pipe in enumerate(self.out_pipes):
+            first_segment_line, *remainder = pipe.export_str(pipe_idx=i).split('\n', maxsplit=1)
+            first_segment_lines.append(first_segment_line)
+            remainders.extend(remainder)
 
-        return f"COMPONENT:'{self.type}',{self.posn[0]},{self.posn[1]},''\n" \
-               + '\n'.join(pipe.export_str(pipe_idx=i) for i, pipe in enumerate(self.out_pipes))
+        return '\n'.join([f"COMPONENT:'{self.type}',{self.posn.col},{self.posn.row},''",
+                          *first_segment_lines, *remainders])
 
 
 class Input(Component):
+    DEFAULT_SHAPE = (2, 3)
     __slots__ = 'molecules', 'input_rate', 'num_inputs'
 
     # Convenience property for when we know we're dealing with an Input
@@ -283,7 +276,7 @@ class Input(Component):
         self.out_pipes[0] = p
 
     def __new__(cls, input_dict, *args, **kwargs):
-        '''Convert to a random or programmed input if relevant.'''
+        """Convert to a random or programmed input if relevant."""
         if 'repeating-molecules' in input_dict:
             return object.__new__(ProgrammedInput)
 
@@ -345,7 +338,7 @@ class RandomInput(Input):
         self.input_counts = [input_mol_dict['count'] for input_mol_dict in input_dict[molecules_key]]
 
     def get_next_molecule_idx(self):
-        '''Get the next input molecule's index. Exposed to allow for tracking branches in random level states.'''
+        """Get the next input molecule's index. Exposed to allow for tracking branches in random level states."""
         # Create the next balance bucket if we've run out.
         # The bucket stores an index identifying one of the 2-3 molecules
         if not self.random_bucket:
@@ -422,6 +415,7 @@ class ProgrammedInput(Input):
 
 
 class Output(Component):
+    DEFAULT_SHAPE = (2, 3)
     __slots__ = 'output_molecule', 'target_count', 'current_count'
 
     # Convenience property for when we know we're dealing with an Output
@@ -445,9 +439,9 @@ class Output(Component):
         self.current_count = 0
 
     def do_instant_actions(self, _):
-        '''Check for and process any incoming molecule, and return True if this output just completed (in which case
+        """Check for and process any incoming molecule, and return True if this output just completed (in which case
         the caller should check if the other outputs are also done). This avoids checking all output counts every cycle.
-        '''
+        """
         if self.in_pipe is None:
             return False
 
@@ -494,9 +488,9 @@ class PassThroughCounter(Output):
         self.out_pipes[0] = p
 
     def do_instant_actions(self, cycle):
-        '''Check for and process any incoming molecule, and return True if this output just completed (in which case
+        """Check for and process any incoming molecule, and return True if this output just completed (in which case
         the caller should check if the other outputs are also done). This avoids checking all output counts every cycle.
-        '''
+        """
         if self.in_pipe is None:
             return False
 
@@ -526,7 +520,8 @@ class PassThroughCounter(Output):
 
 # It's less confusing for output counting and user-facing purposes if this is not an Output subclass
 class DisabledOutput(Component):
-    '''Used by research levels, which actually crash if a wrong output is used unlike assembly reactors.'''
+    """Used by research levels, which actually crash if a wrong output is used unlike assembly reactors."""
+    DEFAULT_SHAPE = (1, 1)
     __slots__ = ()
 
     @property
@@ -545,6 +540,7 @@ class DisabledOutput(Component):
 
 class OutputPrinter(Component):
     """Displays the last 3 molecules passed to it. For now this is effectively going to be a recycler..."""
+    DEFAULT_SHAPE = (2, 3)
     __slots__ = ()
 
     def __init__(self, component_dict=None, _type=None, posn=None):
@@ -583,9 +579,9 @@ class PassThroughPrinter(OutputPrinter):
         self.out_pipes[0] = p
 
     def do_instant_actions(self, cycle):
-        '''Check for and process any incoming molecule, and return True if this output just completed (in which case
+        """Check for and process any incoming molecule, and return True if this output just completed (in which case
         the caller should check if the other outputs are also done). This avoids checking all output counts every cycle.
-        '''
+        """
         if self.in_pipe is None:
             return
 
@@ -606,6 +602,7 @@ class PassThroughPrinter(OutputPrinter):
 
 
 class Recycler(Component):
+    DEFAULT_SHAPE = (5, 5)
     __slots__ = ()
 
     def __init__(self, component_dict=None, _type=None, posn=None):
@@ -620,6 +617,7 @@ class Recycler(Component):
 # TODO: Ideally this would subclass both deque and Component but doing so gives me
 #       "multiple bases have instance lay-out conflict". Need to investigate.
 class StorageTank(Component):
+    DEFAULT_SHAPE = (3, 3)
     MAX_CAPACITY = 25
     __slots__ = 'contents',
 
@@ -687,6 +685,7 @@ class InfiniteStorageTank(StorageTank):
 
 
 class TeleporterInput(Component):
+    DEFAULT_SHAPE = (3, 1)
     __slots__ = 'destination',
 
     def __init__(self, component_dict):
@@ -703,11 +702,11 @@ class TeleporterInput(Component):
         self.in_pipes[0] = p
 
     def do_instant_actions(self, _):
-        '''Note that the teleporter pair behaves differently from a pass-through counter insofar as the pass-through
+        """Note that the teleporter pair behaves differently from a pass-through counter insofar as the pass-through
         counter stores any molecule it receives internally when its output pipe is clogged, whereas the teleporter
         refuses to accept the next molecule until the output pipe is clear (i.e. behaves like a single discontinuous
         pipe that also happens to only allow single atoms through).
-        '''
+        """
         if self.in_pipe is None:
             return
 
@@ -719,6 +718,7 @@ class TeleporterInput(Component):
 
 
 class TeleporterOutput(Component):
+    DEFAULT_SHAPE = (3, 1)
     # TODO: Needing an internal molecule slot is awkward but I couldn't find a cleaner way to avoid the molecule
     #       being both teleported and moved in the same cycle if the teleporters have the wrong relative component
     #       priorities
@@ -750,6 +750,7 @@ class TeleporterOutput(Component):
 
 
 class Reactor(Component):
+    DEFAULT_SHAPE = (4, 4)  # Size in overworld
     # For convenience during float-precision rotation co-ordinates, we consider the center of the
     # top-left cell to be at (0,0), and hence the top-left reactor corner is (-0.5, -0.5).
     # Further, treat the walls as being one atom radius closer, so that we can efficiently check if an atom will collide
@@ -762,18 +763,17 @@ class Reactor(Component):
              LEFT: -0.5 + ATOM_RADIUS, RIGHT: 9.5 - ATOM_RADIUS}
     # Names of features as stored in attributes
     FEATURE_NAMES = ('bonders', 'sensors', 'fusers', 'splitters', 'swappers')
-
     __slots__ = ('in_pipes', 'out_pipes',
                  'waldos', 'molecules',
-                 'large_output', 'bonders', 'sensors', 'fusers', 'splitters', 'swappers',
+                 'large_output', *FEATURE_NAMES,
                  'bond_plus_pairs', 'bond_minus_pairs',
                  'quantum_walls_x', 'quantum_walls_y', 'disallowed_instrs',
-                 'debug')
+                 'annotations', 'debug')
 
     def __init__(self, component_dict=None, _type=None, posn=None):
-        '''Initialize a reactor from only its component dict, doing e.g. default placements of features. Used for
+        """Initialize a reactor from only its component dict, doing e.g. default placements of features. Used for
         levels with preset reactors.
-        '''
+        """
         if component_dict is None:
             component_dict = {}
 
@@ -853,15 +853,16 @@ class Reactor(Component):
                     out_list.append((a, (wall_min - 0.5, wall_max + 0.5)))
 
         self.disallowed_instrs = set() if 'disallowed-instructions' not in component_dict else set(component_dict['disallowed-instructions'])
+        self.annotations = []
 
         # Store molecules as dict keys to be ordered (preserving Spacechem's hidden 'least recently modified' rule)
         # and to have O(1) add/delete. Dict values are ignored.
         self.molecules = {}
 
     def bond_pairs(self):
-        '''For each of + and - bond commands, return a tuple of (bonder_A_posn, bonder_B_posn, dirn) triplets,
+        """For each of + and - bond commands, return a tuple of (bonder_A_posn, bonder_B_posn, dirn) triplets,
         sorted in priority order.
-        '''
+        """
         pair_lists = []
         for bond_type in ('+', '-'):
             # Store the relevant types of bonders in a dict paired up with their indices for fast lookup/sorting (below)
@@ -888,17 +889,20 @@ class Reactor(Component):
         # Break the component string up into its individual sections, while removing empty lines
         component_line, *lines = (s for s in export_str.split('\n') if s)
 
+        # Member lines
         pipes_idx = next((i for i, s in enumerate(lines) if s.startswith('PIPE:')), len(lines))
         member_lines, lines = lines[:pipes_idx], lines[pipes_idx:]
         if not member_lines:
             raise ValueError("Missing MEMBER lines in reactor component")
 
+        # Pipe and annotation lines
         annotations_idx = next((i for i, s in enumerate(lines) if s.startswith('ANNOTATION:')), len(lines))
-        pipes_str = '\n'.join(lines[:annotations_idx])
+        pipe_lines, annotation_lines = lines[:annotations_idx], lines[annotations_idx:]
 
         # Validates COMPONENT line and updates pipes
-        super().update_from_export_str(component_line + '\n' + pipes_str, update_pipes=update_pipes)
+        super().update_from_export_str(component_line + '\n' + '\n'.join(pipe_lines), update_pipes=update_pipes)
 
+        # Add members (features and instructions)
         for line in member_lines:
             assert line.startswith('MEMBER:'), f"Unexpected line in reactor members: `{line}`"
             fields = line.split(',')
@@ -1039,13 +1043,26 @@ class Reactor(Component):
                 f"Expected {len(getattr(self, feature_name))} {feature_name} for {self.type} reactor but got {len(posns)}"
             setattr(self, feature_name, posns)
 
-        self.bond_plus_pairs, self.bond_minus_pairs = self.bond_pairs()  # Re-precompute bond pairings based on the changes
+        self.bond_plus_pairs, self.bond_minus_pairs = self.bond_pairs()  # Re-precompute bond pairings
+
+        # Store the annotations sorted by output idx (the first field)
+        self.annotations = sorted(annotation_lines)
 
     def export_str(self):
-        '''Represent this reactor in solution export string format.'''
-        export_str = f"COMPONENT:'{self.type}',{self.posn.col},{self.posn.row},''"
+        """Represent this reactor in solution export string format."""
+        # Generate the generic component export, and separate it into the metadata and pipe lines
+        component_line, *pipes = super().export_str().split('\n', maxsplit=1)
+
+        # By SC's convention, MEMBER lines are ordered as waldo starts, then features, then remaining waldo instructions
+        # Grab the waldo start lines from the front of each waldo's export
+        waldo_starts, waldo_instrs = [], []
+        for waldo in self.waldos:
+            start, *instrs = waldo.export_str().split('\n', maxsplit=1)
+            waldo_starts.append(start)
+            waldo_instrs.extend(instrs)
 
         # TODO: Make reactors more agnostic of feature types
+        features = []
         for (posn, bond_types) in self.bonders:
             if bond_types == '+-':
                 feature_name = 'bonder'
@@ -1056,28 +1073,25 @@ class Reactor(Component):
             else:
                 raise Exception("Invalid bonder type in internal data")
 
-            export_str += f"\nMEMBER:'feature-{feature_name}',-1,0,1,{posn.col},{posn.row},0,0"
+            features.append(f"MEMBER:'feature-{feature_name}',-1,0,1,{posn.col},{posn.row},0,0")
         for posn in self.sensors:
-            export_str += f"\nMEMBER:'feature-sensor',-1,0,1,{posn.col},{posn.row},0,0"
+            features.append(f"MEMBER:'feature-sensor',-1,0,1,{posn.col},{posn.row},0,0")
         for posn in self.fusers:
-            export_str += f"\nMEMBER:'feature-fuser',-1,0,1,{posn.col},{posn.row},0,0"
+            features.append(f"MEMBER:'feature-fuser',-1,0,1,{posn.col},{posn.row},0,0")
         for posn in self.splitters:
-            export_str += f"\nMEMBER:'feature-splitter',-1,0,1,{posn.col},{posn.row},0,0"
+            features.append(f"MEMBER:'feature-splitter',-1,0,1,{posn.col},{posn.row},0,0")
         for posn in self.swappers:
-            export_str += f"\nMEMBER:'feature-tunnel',-1,0,1,{posn.col},{posn.row},0,0"
+            features.append(f"MEMBER:'feature-tunnel',-1,0,1,{posn.col},{posn.row},0,0")
 
-        export_str += '\n' + '\n'.join(waldo.export_str() for waldo in self.waldos)
-        export_str += '\n' + '\n'.join(pipe.export_str(pipe_idx=i) for i, pipe in enumerate(self.out_pipes))
-
-        return export_str
+        return '\n'.join([component_line, *waldo_starts, *features, *waldo_instrs, *pipes, *self.annotations])
 
     def __hash__(self):
-        '''Hash of the current reactor state.'''
+        """Hash of the current reactor state."""
         return hash((tuple(molecule.hashable_repr() for molecule in self.molecules),
                      tuple(self.waldos)))
 
     def __str__(self, flash_features=True, show_instructions=False):
-        '''Return a rich-format pretty-print string representing this reactor.'''
+        """Return a rich-format pretty-print string representing this reactor."""
         # Each cell gets two characters, + 1 space between cells (we'll use that space to show waldos reticles)
         cells = [[[' ', ' '] for _ in range(self.NUM_COLS)] for _ in range(self.NUM_ROWS)]
         borders = [[' ' for _ in range(self.NUM_COLS + 1)] for _ in range(self.NUM_ROWS)]  # Waldos and zone edges
@@ -1244,7 +1258,7 @@ class Reactor(Component):
             self.exec_instrs(waldo)
 
     def move_contents(self, cycle):
-        '''Move all waldos in this reactor and any molecules they are holding.'''
+        """Move all waldos in this reactor and any molecules they are holding."""
         super().move_contents(cycle)
 
         # If the waldo is facing a wall, mark it as stalled (may also be stalled due to sync, input, etc.)
@@ -1353,14 +1367,14 @@ class Reactor(Component):
             waldo.is_stalled = False
 
     def check_molecule_collisions_lazy(self, molecule):
-        '''Raise an exception if the given molecule collides with any other molecules.
+        """Raise an exception if the given molecule collides with any other molecules.
         Assumes integer co-ordinates in all molecules.
-        '''
+        """
         for other_molecule in self.molecules:
             molecule.check_collisions_lazy(other_molecule)  # Implicitly ignores self
 
     def check_wall_collisions(self, molecule):
-        '''Raise an exception if the given molecule collides with any walls.'''
+        """Raise an exception if the given molecule collides with any walls."""
         if not all(self.walls[UP] < p.row < self.walls[DOWN]
                    and self.walls[LEFT] < p.col < self.walls[RIGHT]
                    for p in molecule.atom_map):
@@ -1388,17 +1402,17 @@ class Reactor(Component):
                     raise ReactionError("A molecule has collided with a quantum wall")
 
     def check_collisions_lazy(self, molecule):
-        '''Raise an exception if the given molecule collides with any other molecules or walls.
+        """Raise an exception if the given molecule collides with any other molecules or walls.
         Assumes integer co-ordinates in all molecules.
-        '''
+        """
         self.check_molecule_collisions_lazy(molecule)
         self.check_wall_collisions(molecule)
         # Quantum wall collision checks may be skipped since they should only lie on grid edges
 
     def check_collisions(self, molecule):
-        '''Check that the given molecule isn't colliding with any walls or other molecules.
+        """Check that the given molecule isn't colliding with any walls or other molecules.
         Raise an exception if it is.
-        '''
+        """
         for other_molecule in self.molecules:
             molecule.check_collisions(other_molecule)  # Implicitly ignores self
 
@@ -1528,9 +1542,9 @@ class Reactor(Component):
         waldo.is_stalled = molecule is not None
 
     def get_molecule(self, position):
-        '''Select the molecule at the given grid position, or None if no such molecule.
+        """Select the molecule at the given grid position, or None if no such molecule.
         Used by Grab, Bond+/-, Fuse, etc.
-        '''
+        """
         return next((molecule for molecule in self.molecules if position in molecule), None)
 
     def grab(self, waldo):
@@ -1622,11 +1636,11 @@ class Reactor(Component):
                         waldo.molecule = split_off_molecule
 
     def defrag_molecule(self, molecule, posn):
-        '''Given a molecule that has had some of its bonds broken from the given position, update reactor.molecules
+        """Given a molecule that has had some of its bonds broken from the given position, update reactor.molecules
         based on any molecules that broke off. Note that this always at least moves the molecule to the back of the
         priority queue, even if it did not break apart (this should be safe since defrag should only be called when the
         molecule is modified).
-        '''
+        """
         # Update the reactor molecules based on how the molecule broke apart
         del self.molecules[molecule]
         for new_molecule in molecule.defrag(posn):
@@ -1638,9 +1652,9 @@ class Reactor(Component):
                     waldo.molecule = new_molecule
 
     def delete_atom_bonds(self, posn):
-        '''Helper used by fuse and swap to remove all bonds from an atom and break up its molecule if needed.
+        """Helper used by fuse and swap to remove all bonds from an atom and break up its molecule if needed.
         If no atom at the given position, does nothing.
-        '''
+        """
         molecule = self.get_molecule(posn)
         if molecule is None:
             return
@@ -1655,9 +1669,9 @@ class Reactor(Component):
         self.defrag_molecule(molecule, posn)
 
     def reduce_excess_bonds(self, posn):
-        '''Helper used by fuse and split to reduce bonds on a mutated atom down to its new max count, and break up its
+        """Helper used by fuse and split to reduce bonds on a mutated atom down to its new max count, and break up its
         molecule if needed.
-        '''
+        """
         molecule = self.get_molecule(posn)
         atom = molecule.atom_map[posn]
 
@@ -1750,9 +1764,9 @@ class Reactor(Component):
             self.molecules[new_molecule] = None  # Dummy value
 
     def swap(self):
-        '''Swap atoms between swappers. Note that the order of operations here was carefully chosen to modify the
+        """Swap atoms between swappers. Note that the order of operations here was carefully chosen to modify the
         internal priority order of reactor molecules the same way that SpaceChem does.
-        '''
+        """
         # Debond all atoms on swappers from their neighbors
         for posn in self.swappers:
             self.delete_atom_bonds(posn)  # Does nothing if no atom on the swapper
